# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FBOutput

import tdw.flatbuffers

class SubObject(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsSubObject(cls, buf, offset):
        n = tdw.flatbuffers.encode.Get(tdw.flatbuffers.packer.uoffset, buf, offset)
        x = SubObject()
        x.Init(buf, n + offset)
        return x

    # SubObject
    def Init(self, buf, pos):
        self._tab = tdw.flatbuffers.table.Table(buf, pos)

    # SubObject
    def Id(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(tdw.flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # SubObject
    def MachineType(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(tdw.flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 1

def SubObjectStart(builder): builder.StartObject(2)
def SubObjectAddId(builder, id): builder.PrependInt32Slot(0, id, 0)
def SubObjectAddMachineType(builder, machineType): builder.PrependUint8Slot(1, machineType, 1)
def SubObjectEnd(builder): return builder.EndObject()
