# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FBOutput

import tdw.flatbuffers

class AudioSource(object):
    __slots__ = ['_tab']

    # AudioSource
    def Init(self, buf, pos):
        self._tab = tdw.flatbuffers.table.Table(buf, pos)

    # AudioSource
    def Id(self): return self._tab.Get(tdw.flatbuffers.number_types.Int32Flags, self._tab.Pos + tdw.flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # AudioSource
    def IsPlaying(self): return self._tab.Get(tdw.flatbuffers.number_types.BoolFlags, self._tab.Pos + tdw.flatbuffers.number_types.UOffsetTFlags.py_type(4))

def CreateAudioSource(builder, id, isPlaying):
    builder.Prep(4, 8)
    builder.Pad(3)
    builder.PrependBool(isPlaying)
    builder.PrependInt32(id)
    return builder.Offset()
